import inspect
import os

from connect.eaas.core.extension import TransformationsApplicationBase
from connect.eaas.core.validation.helpers import get_code_context
from connect.eaas.core.validation.models import ValidationItem, ValidationResult


def validate_tfnapp(context):  # noqa: CCR001
    messages = []

    if 'tfnapp' not in context['extension_classes']:
        return ValidationResult()

    extension_class = context['extension_classes']['tfnapp']
    extension_class_file = inspect.getsourcefile(extension_class)

    if not issubclass(extension_class, TransformationsApplicationBase):
        messages.append(
            ValidationItem(
                level='ERROR',
                message=(
                    f'The application class *{extension_class.__name__}* '
                    f'is not a subclass of '
                    f'*connect.eaas.core.extension.TransformationsApplicationBase*.'
                ),
                file=extension_class_file,
            ),
        )
        return ValidationResult(items=messages, must_exit=True)

    transformations = extension_class.get_transformations()
    for tfn in transformations:
        if not isinstance(tfn['name'], str) or not tfn['name']:
            messages.append(
                ValidationItem(
                    level='ERROR',
                    message=(
                        f'The name of the {tfn["method"]} must be a non-blank string.'
                    ),
                    **get_code_context(extension_class, tfn['method']),
                ),
            )
        if not isinstance(tfn['description'], str) or not tfn['description']:
            messages.append(
                ValidationItem(
                    level='ERROR',
                    message=(
                        f'The description of the {tfn["method"]} must be a non-blank string.'
                    ),
                    **get_code_context(extension_class, tfn['method']),
                ),
            )

        edit_dialog_ui = tfn['edit_dialog_ui']
        if (
            not isinstance(edit_dialog_ui, str)
            or not edit_dialog_ui
            or not edit_dialog_ui.startswith('/static/')
        ):
            messages.append(
                ValidationItem(
                    level='ERROR',
                    message=(
                        f'The url {edit_dialog_ui} of the {tfn["method"]} '
                        f'must be not empty string and prefixed with /static.'
                    ),
                    **get_code_context(extension_class, tfn['method']),
                ),
            )
        else:
            full_path = os.path.join(
                os.path.dirname(inspect.getsourcefile(extension_class)),
                'static',
                edit_dialog_ui[8:],
            )
            if not os.path.exists(full_path):
                messages.append(
                    ValidationItem(
                        level='ERROR',
                        message=(
                            f'The url {edit_dialog_ui} of the {tfn["method"]} '
                            f'does not point to any file.'
                        ),
                        **get_code_context(extension_class, tfn['method']),
                    ),
                )

    return ValidationResult(items=messages)



import re
import subprocess
from collections import OrderedDict
from distutils.version import StrictVersion


DEFAULT_CLONE_DIR = '_repo'


class GitException(Exception):
    pass


def clone_repo(temp_path, repo_url, tag):
    result = subprocess.run(
        [
            'git',
            '-C',
            temp_path,
            'clone',
            '-b',
            tag,
            repo_url,
            DEFAULT_CLONE_DIR,
        ],
        capture_output=True,
        stdin=subprocess.DEVNULL,
        start_new_session=True,
    )
    try:
        result.check_returncode()
    except subprocess.CalledProcessError:
        raise GitException(f'Error cloning repository {repo_url}: {result.stderr.decode()}')


class ConnectVersionTag(StrictVersion):
    version_re = re.compile(
        r'^v?(\d+) \. (\d+) (\. (\d+))? ([ab](\d+))?$',
        re.VERBOSE | re.ASCII,
    )

    plain_tag = None

    def parse(self, vstring):
        try:
            super().parse(vstring)
        except Exception:
            self.plain_tag = vstring

    def _cmp_plain_tag(self, actual, other):
        if actual and other:
            if actual == other:
                return 0
            return -1 if actual > other else 1
        elif not actual and other:
            return 1
        if actual and not other:
            return -1
        return None

    def _cmp(self, other):
        result = self._cmp_plain_tag(self.plain_tag, other.plain_tag)
        if result is not None:
            return result

        return super()._cmp(other)


def sort_tags(tags):
    sorted_tags = OrderedDict()
    for tag in sorted(tags.keys(), key=ConnectVersionTag, reverse=True):
        sorted_tags[tag] = tags[tag]

    return sorted_tags


def list_tags(repo_url):
    version_re = re.compile(
        r'^v?(\d+) \. (\d+) (\. (\d+))?$',
        re.VERBOSE | re.ASCII,
    )

    result = subprocess.run(
        ['git', 'ls-remote', '--tags', '--refs', repo_url],
        capture_output=True,
        stdin=subprocess.DEVNULL,
        start_new_session=True,
    )
    try:
        result.check_returncode()
    except subprocess.CalledProcessError:
        raise GitException(result.stderr.decode())

    tags = {}
    for line in result.stdout.decode().splitlines():
        commit, tagref = line.split()
        tag = tagref.rsplit('/', 1)[-1]
        if version_re.match(tag):
            tags[tag] = commit
    return sort_tags(tags)



from connect.client import ClientError
from connect.eaas.core.deployment.utils import (
    DeploymentError,
    create_extension,
    extract_arguments,
    get_git_data,
    process_variables,
    stop_environment,
    update_extension,
)


def deploy_extension(repo, client, log, tag=None):  # noqa: CCR001
    log(f'Starting deploy repo {repo}...')

    try:
        log('Getting tags information.')
        git = get_git_data(repo, tag)
        log(f'Tag {git["tag"]} is set.')

        log('Extracting data from deployment file.')
        arguments = extract_arguments(repo, git['tag'])
    except DeploymentError as de:
        log(de)
        return

    if not arguments.get('package_id'):
        log('No required package_id found.')
        return
    arguments['name'] = arguments.get('name') or f'Extension for {arguments["package_id"]}'

    try:
        log(f'Looking up for extension with package_id {arguments["package_id"]}.')
        extension = client('devops').services.filter(package_id=arguments['package_id']).first()
    except ClientError as ce:
        log(f'Error getting extension: {ce}')
        return

    log(
        (
            f'Extension with package_id {arguments["package_id"]} '
            f'does{"" if extension else " not"} exist, '
            f'{"updating" if extension else "creating"} it.'
        ),
    )
    is_install = extension is None

    if is_install:
        extension = create_extension(client, arguments, log)
    else:
        extension = update_extension(client, extension, arguments, log)

    if arguments.get('icon'):
        arguments['icon'].close()
    if extension is None:
        return

    environment = extension['environments'][arguments.get('env', 'production')]
    env_api = client('devops').services[extension['id']].environments[environment['id']]

    if environment['runtime'] == 'local' and environment['status'] == 'connected':
        log(
            (
                f'{environment["type"]} environment is in local mode and connected. '
                f'Please, stop it and rerun deployment again.'
            ),
        )
        return

    try:
        log(f'Processing variables for extension with package_id {arguments["package_id"]}.')
        vars_updated = process_variables(arguments, env_api, log)

        if (
            environment.get('git', {}).get('commit') != git['commit']
            or environment['runtime'] == 'local'
        ):
            environment = stop_environment(environment, env_api, log)
            if environment['status'] in ['running', 'stopping']:
                return

            log(f'Updating {environment["type"]} environment.')
            environment = env_api.update({'git': git, 'runtime': 'cloud'})
        else:
            log(
                (
                    f'Version {git["tag"]} is already deployed in cloud mode for '
                    f'{environment["type"]} environment for extension with package_id '
                    f'{arguments["package_id"]}.'
                ),
            )

        if environment['status'] not in ['running', 'deploying']:
            log(f'Starting {environment["type"]} environment.')
            env_api.action('start').post()
        elif environment['status'] == 'running' and vars_updated:
            log(f'Updating {environment["type"]} environment config.')
            env_api.action('update-config').post()
    except ClientError as ce:
        log(f'Error processing environment: {ce}.')
        return

    log(f'Extension with package_id {arguments["package_id"]} successfully deployed.')



#
# Copyright (c) 2023 Ingram Micro. All Rights Reserved.
#

import inspect
import os

from fastapi_utils.cbv import CBV_CLASS_KEY

from connect.eaas.core.constants import (
    PROXIED_CONNECT_API_ENDPOINTS_MAX_ALLOWED_NUMBER,
    PROXIED_CONNECT_API_ENDPOINTS_PUBLIC_PREFIX,
)
from connect.eaas.core.extension import WebApplicationBase
from connect.eaas.core.validation.helpers import get_code_context
from connect.eaas.core.validation.models import ValidationItem, ValidationResult


def validate_webapp(context):  # noqa: CCR001

    messages = []

    if 'webapp' not in context['extension_classes']:
        return ValidationResult()

    extension_class = _get_extension_class(context)
    extension_class_file = _get_extension_class_file(extension_class)

    if not issubclass(extension_class, WebApplicationBase):
        messages.append(
            ValidationItem(
                level='ERROR',
                message=(
                    f'The application class *{extension_class.__name__}* '
                    f'is not a subclass of *connect.eaas.core.extension.WebApplicationBase*.'
                ),
                file=extension_class_file,
            ),
        )
        return ValidationResult(items=messages, must_exit=True)

    if not getattr(extension_class, CBV_CLASS_KEY, False):
        messages.append(
            ValidationItem(
                level='ERROR',
                message='The Web application class must be wrapped in *@web_app(router)*.',
                file=extension_class_file,
            ),
        )
        return ValidationResult(items=messages, must_exit=True)

    validators = [
        _validate_webapp_exception_handling,
        _validate_webapp_middlewares,
        _validate_webapp_routes,
        _validate_webapp_ui_modules,
        _validate_webapp_proxied_connect_api,
    ]
    for validator in validators:
        messages.extend(validator(context))

    return ValidationResult(items=messages, must_exit=len(messages) > 0)


def _validate_webapp_exception_handling(context):
    errors = []

    extension_class = _get_extension_class(context)
    if not hasattr(extension_class, 'get_exception_handlers'):
        return errors

    extension_class_file = _get_extension_class_file(extension_class)
    try:
        exceptions_dict = extension_class.get_exception_handlers({})
        if not isinstance(exceptions_dict, dict):
            errors.append(
                ValidationItem(
                    level='ERROR',
                    message=(
                        'Invalid implementation of `get_exception_handlers` in Web Application: '
                        'returned value must be a dictionary.'
                    ),
                    file=extension_class_file,
                ),
            )
            return errors

        if not all(issubclass(k, BaseException) for k in exceptions_dict.keys()):
            errors.append(
                ValidationItem(
                    level='ERROR',
                    message=(
                        'Invalid implementation of `get_exception_handlers` in Web Application: all'
                        ' configuration keys must Exception classes, inherited from BaseException.'
                    ),
                    file=extension_class_file,
                ),
            )
            return errors

        return errors

    except TypeError:
        errors.append(
            ValidationItem(
                level='ERROR',
                message='Invalid declaration of `get_exception_handlers` in Web Application.',
                file=extension_class_file,
            ),
        )
        return errors


def _validate_webapp_middlewares(context):
    errors = []

    extension_class = _get_extension_class(context)
    if not hasattr(extension_class, 'get_middlewares'):
        return errors

    extension_class_file = _get_extension_class_file(extension_class)
    try:
        middlewares_list = extension_class.get_middlewares()
        if not isinstance(middlewares_list, list):
            errors.append(
                ValidationItem(
                    level='ERROR',
                    message=(
                        'Invalid implementation of `get_middlewares` in Web Application: '
                        'returned value must be a list.'
                    ),
                    file=extension_class_file,
                ),
            )
            return errors

        return errors

    except TypeError:
        errors.append(
            ValidationItem(
                level='ERROR',
                message='Invalid declaration of `get_middlewares` in Web Application.',
                file=extension_class_file,
            ),
        )
        return errors


def _validate_webapp_proxied_connect_api(context):
    errors = []

    extension_class = _get_extension_class(context)
    proxied_connect_api = extension_class.get_proxied_connect_api()
    code_pattern = '@proxied_connect_api('

    type_msg = 'The argument of the `@proxied_connect_api` must be a list of strings or a dict.'
    if not isinstance(proxied_connect_api, (list, dict)):
        errors.append(
            ValidationItem(
                level='ERROR',
                message=type_msg,
                **get_code_context(extension_class, code_pattern),
            ),
        )
        return errors

    if not proxied_connect_api:
        return errors

    max_length = PROXIED_CONNECT_API_ENDPOINTS_MAX_ALLOWED_NUMBER
    if len(proxied_connect_api) > max_length:
        errors.append(
            ValidationItem(
                level='ERROR',
                message=f'Max allowed length of the `@proxied_connect_api` argument: {max_length}.',
                **get_code_context(extension_class, code_pattern),
            ),
        )
        return errors

    is_dict_conf = isinstance(proxied_connect_api, dict)
    api_paths = proxied_connect_api
    if is_dict_conf:
        api_paths = proxied_connect_api.keys()
    if not all(isinstance(v, str) for v in api_paths):
        errors.append(
            ValidationItem(
                level='ERROR',
                message=type_msg,
                **get_code_context(extension_class, code_pattern),
            ),
        )
        return errors

    public_prefix = PROXIED_CONNECT_API_ENDPOINTS_PUBLIC_PREFIX
    if not all(v.startswith(public_prefix) for v in api_paths):
        errors.append(
            ValidationItem(
                level='ERROR',
                message='Only Public API can be referenced in `@proxied_connect_api`.',
                **get_code_context(extension_class, code_pattern),
            ),
        )
        return errors

    if is_dict_conf and (
        not all(isinstance(v, str) and v in ('view', 'edit') for v in proxied_connect_api.values())
    ):
        errors.append(
            ValidationItem(
                level='ERROR',
                message='Wrong Public API permission value in `@proxied_connect_api`.',
                **get_code_context(extension_class, code_pattern),
            ),
        )
        return errors

    return errors


def _validate_webapp_routes(context):
    extension_class = _get_extension_class(context)
    extension_class_file = _get_extension_class_file(extension_class)

    auth, no_auth = extension_class.get_routers()

    no_of_routes = len(auth.routes) + len(no_auth.routes)

    if no_of_routes == 0:
        return [
            ValidationItem(
                level='ERROR',
                message=(
                    'The Web application class must contain at least one route '
                    'implementation function wrapped in *@router.your_method("/your_path")*.'
                ),
                file=extension_class_file,
            ),
        ]
    return []


def _check_ui_component_label(extension_class, page_name, value, code_pattern):
    if not isinstance(value, str) or not value:
        return [
            ValidationItem(
                level='ERROR',
                message=(
                    f'The label of the {page_name} must be a non-blank string.'
                ),
                **get_code_context(extension_class, code_pattern),
            ),
        ]
    return []


def _generate_ui_component_error(extension_class, parameter_name, format, page_name, code_pattern):
    return [
        ValidationItem(
            level='ERROR',
            message=(
                f'The {parameter_name} of the {page_name} must be a '
                f'path to {format} file relative to the static folder.'
            ),
            **get_code_context(extension_class, code_pattern),
        ),
    ]


def _generate_ui_component_static(extension_class, parameter_name, page_name, value, code_pattern):
    return [
        ValidationItem(
            level='ERROR',
            message=(
                f'The {parameter_name} {value} of the {page_name} must be prefixed with /static.'
            ),
            **get_code_context(extension_class, code_pattern),
        ),
    ]


def _generate_ui_component_does_not_exist(
    extension_class,
    parameter_name,
    page_name,
    value,
    code_pattern,
):
    return [
        ValidationItem(
            level='ERROR',
            message=(
                f'The {parameter_name} {value} of the {page_name} page does not point to any file.'
            ),
            **get_code_context(extension_class, code_pattern),
        ),
    ]


def _check_ui_component_url(extension_class, page_name, value, code_pattern):
    if not isinstance(value, str) or not value:
        return _generate_ui_component_error(extension_class, 'url', 'html', page_name, code_pattern)

    if not value.startswith('/static/'):
        return _generate_ui_component_static(extension_class, 'url', page_name, value, code_pattern)

    page_path = value[8:]
    full_path = os.path.join(
        os.path.dirname(_get_extension_class_file(extension_class)),
        'static',
        page_path,
    )
    if not os.path.exists(full_path):
        return _generate_ui_component_does_not_exist(
            extension_class,
            'url',
            page_name,
            value,
            code_pattern,
        )

    return []


def _check_ui_component_icon(extension_class, page_name, value, code_pattern):
    if not isinstance(value, str) or not value:
        return _generate_ui_component_error(
            extension_class,
            'icon',
            'image',
            page_name,
            code_pattern,
        )

    if not value.startswith('/static/'):
        return _generate_ui_component_static(
            extension_class,
            'icon',
            page_name,
            value,
            code_pattern,
        )

    page_path = value[8:]
    full_path = os.path.join(
        os.path.dirname(_get_extension_class_file(extension_class)),
        'static',
        page_path,
    )
    if not os.path.exists(full_path):
        return _generate_ui_component_does_not_exist(
            extension_class,
            'icon',
            page_name,
            value,
            code_pattern,
        )

    return []


def _validate_webapp_ui_modules(context):  # noqa: CCR001
    messages = []
    extension_class = _get_extension_class(context)

    ui_modules = extension_class.get_ui_modules()

    if 'settings' in ui_modules:
        label = ui_modules['settings']['label']
        url = ui_modules['settings']['url']

        messages.extend(
            _check_ui_component_label(
                extension_class, '"Account Settings"', label, '@account_settings_page(',
            ),
        )

        messages.extend(
            _check_ui_component_url(
                extension_class, '"Account Settings"', url, '@account_settings_page(',
            ),
        )

    if 'modules' in ui_modules:
        label = ui_modules['modules']['label']
        url = ui_modules['modules']['url']
        messages.extend(
            _check_ui_component_label(
                extension_class, '"Module Main Page"', label, '@module_pages(',
            ),
        )
        messages.extend(
            _check_ui_component_url(
                extension_class, '"Module Main Page"', url, '@module_pages(',
            ),
        )

        if 'children' in ui_modules['modules']:
            children = ui_modules['modules']['children']
            if not isinstance(children, (list, tuple)):
                messages.append(
                    ValidationItem(
                        level='ERROR',
                        message=(
                            'The "children" argument of the @module_pages '
                            'decorator must be a list of objects.'
                        ),
                        **get_code_context(extension_class, '@module_pages('),
                    ),
                )
            else:
                for child in children:
                    if 'label' not in child or 'url' not in child:
                        messages.append(
                            ValidationItem(
                                level='ERROR',
                                message=(
                                    'Invalid child page declaration. '
                                    'Each child page must be an object with the '
                                    'label and url attributes.'
                                ),
                                **get_code_context(extension_class, '@module_pages('),
                            ),
                        )
                        continue

                    label = child['label']
                    url = child['url']
                    messages.extend(
                        _check_ui_component_label(
                            extension_class, '"Module Child Page"', label, '@module_pages(',
                        ),
                    )
                    messages.extend(
                        _check_ui_component_url(
                            extension_class, '"Module Child Page"', url, '@module_pages(',
                        ),
                    )

    if 'admins' in ui_modules:
        admins = ui_modules['admins']
        if not isinstance(admins, (list, tuple)):
            messages.append(
                ValidationItem(
                    level='ERROR',
                    message=(
                        'The argument of the @admin_pages '
                        'decorator must be a list of objects.'
                    ),
                    **get_code_context(extension_class, '@admin_pages('),
                ),
            )
        else:
            for child in admins:
                if 'label' not in child or 'url' not in child:
                    messages.append(
                        ValidationItem(
                            level='ERROR',
                            message=(
                                'Invalid admin page declaration. '
                                'Each admin page must be an object with the '
                                'label and url attributes.'
                            ),
                            **get_code_context(extension_class, '@admin_pages('),
                        ),
                    )
                    continue

                label = child['label']
                url = child['url']
                messages.extend(
                    _check_ui_component_label(
                        extension_class, '"Admin Page"', label, '@admin_pages(',
                    ),
                )
                messages.extend(
                    _check_ui_component_url(
                        extension_class, '"Admin Page"', url, '@admin_pages(',
                    ),
                )

    if 'devops' in ui_modules:
        devops = ui_modules['devops']
        if not isinstance(devops, (list, tuple)):
            messages.append(
                ValidationItem(
                    level='ERROR',
                    message=(
                        'The argument of the @devops_pages '
                        'decorator must be a list of objects.'
                    ),
                    **get_code_context(extension_class, '@devops_pages('),
                ),
            )
        else:
            for child in devops:
                if 'label' not in child or 'url' not in child:
                    messages.append(
                        ValidationItem(
                            level='ERROR',
                            message=(
                                'Invalid devops page declaration. '
                                'Each devops page must be an object with the '
                                'label and url attributes.'
                            ),
                            **get_code_context(extension_class, '@devops_pages('),
                        ),
                    )
                    continue

                label = child['label']
                url = child['url']
                messages.extend(
                    _check_ui_component_label(
                        extension_class, '"Devops Page"', label, '@devops_pages(',
                    ),
                )
                messages.extend(
                    _check_ui_component_url(
                        extension_class, '"Devops Page"', url, '@devops_pages(',
                    ),
                )

    if 'customer' in ui_modules:
        customer = ui_modules['customer']
        if not isinstance(customer, (list, tuple)):
            messages.append(
                ValidationItem(
                    level='ERROR',
                    message=(
                        'The argument of the @customer_pages '
                        'decorator must be a list of objects.'
                    ),
                    **get_code_context(extension_class, '@customer_pages('),
                ),
            )
        else:
            for child in customer:
                if 'label' not in child or 'url' not in child:
                    messages.append(
                        ValidationItem(
                            level='ERROR',
                            message=(
                                'Invalid customer page declaration. '
                                'Each customer page must be an object with the '
                                'label and url attributes (and optionally icon).'
                            ),
                            **get_code_context(extension_class, '@customer_pages('),
                        ),
                    )
                    continue

                label = child['label']
                url = child['url']
                messages.extend(
                    _check_ui_component_label(
                        extension_class, '"Customer Page"', label, '@customer_pages(',
                    ),
                )
                messages.extend(
                    _check_ui_component_url(
                        extension_class, '"Customer Page"', url, '@customer_pages(',
                    ),
                )
                if 'icon' in child:
                    messages.extend(
                        _check_ui_component_icon(
                            extension_class, '"Customer Page"', child['icon'], '@customer_pages(',
                        ),
                    )

    return messages


def _get_extension_class(context):
    return context['extension_classes']['webapp']


def _get_extension_class_file(extension_class):
    return inspect.getsourcefile(extension_class)



VARIABLES_INFO_ATTR_NAME = '_eaas_variable_info'
EVENT_INFO_ATTR_NAME = '_eaas_event_info'
SCHEDULABLE_INFO_ATTR_NAME = '_eaas_schedulable_info'
ANVIL_KEY_VAR_ATTR_NAME = '_eaas_anvil_key_var'
ANVIL_CALLABLE_ATTR_NAME = '_eaas_anvil_callable'
UNAUTHORIZED_ENDPOINT_ATTR_NAME = '_eaas_unauthorized_endpoint'
ACCOUNT_SETTINGS_PAGE_ATTR_NAME = '_eaas_account_settings_page'
MODULE_PAGES_ATTR_NAME = '_eaas_module_pages'
ADMIN_PAGES_ATTR_NAME = '_eaas_admin_pages'
TRANSFORMATION_ATTR_NAME = '_eaas_transformation_info'
DEVOPS_PAGES_ATTR_NAME = '_eaas_devops_pages'
PROXIED_CONNECT_API_ATTR_NAME = '_eaas_proxied_connect_api'
CUSTOMER_PAGES_ATTR_NAME = '_eaas_customer_pages'


PROXIED_CONNECT_API_ENDPOINTS_MAX_ALLOWED_NUMBER = 100
PROXIED_CONNECT_API_ENDPOINTS_PUBLIC_PREFIX = '/public/'



import inspect
import json
import os
from unittest.mock import patch
from urllib.parse import urlparse

import httpx
from fastapi import FastAPI
from fastapi.params import Depends
from fastapi.staticfiles import StaticFiles
from starlette.middleware.base import BaseHTTPMiddleware
from starlette.routing import Match
from starlette.testclient import TestClient

from connect.client import ClientError
from connect.client.testing import AsyncConnectClientMocker, ConnectClientMocker
from connect.eaas.core.inject import asynchronous, synchronous
from connect.eaas.core.inject.models import Context
from connect.eaas.core.utils import client_error_exception_handler


class WebAppTestClient(TestClient):
    """
    Create a new instance of the WebAppTestClient.
    The WebAppTestClient is based on the python requests library
    and will internally create a requests.Session.
    Each interface method (get, post, ...) accepts the following extra
        keyword arguments:

    * **context**: an instance of connect.eaas.core.inject.models.Context
        to provide the endpoint with a specific call context or a dict that will merged
        with a default context.
        If this kwarg is omitted a default context is provided.
    * **installation**: an installation object to receive inside the called endpoint,
        defaults to None.
    * **config**: a dictionary with key/value pairs representing environment variables.
    * **log_level**: allow to simulate a specific log level for such endpoint call
        default to INFO.


    Usage:

    ```py3
    from connect.eaas.core.testing.testclient import WebAppTestClient

    from myext.webapp import MyWebApplication

    client = WebAppTestClient(MyWebApplication)
    response = client.get('/my/endpoint')
    ```

    Args:
        webapp: The web application class.
        base_url (str): The base url to be used.
    """
    def __init__(self, webapp, base_url='https://example.org/public/v1'):
        self._url_prefix = base_url
        self._webapp_class = webapp
        self._app = self._get_application()
        super().__init__(app=self._app, base_url=base_url)

        self.headers = {
            'X-Connect-Api-Gateway-Url': self._url_prefix,
            'X-Connect-User-Agent': 'eaas-test-client',
            'X-Connect-Installation-Api-Key': 'ApiKey XXXX',
        }

    def request(
        self,
        method,
        url,
        content=None,
        data=None,
        files=None,
        json=None,
        params=None,
        headers=None,
        cookies=None,
        auth=httpx._client.USE_CLIENT_DEFAULT,
        follow_redirects=None,
        allow_redirects=None,
        timeout=httpx._client.USE_CLIENT_DEFAULT,
        extensions=None,
        context=None,
        installation=None,
        config=None,
        log_level=None,
    ):
        if not context:
            context = Context(**self._generate_call_context(installation))
        elif isinstance(context, dict):
            ctx_data = self._generate_call_context(installation)
            ctx_data.update(context)
            context = Context(**ctx_data)

        headers = self._populate_internal_headers(
            headers or {},
            context=context,
            installation=installation,
            config=config,
            log_level=log_level,
        )

        handler, path_params = self._get_endpoint_handler_and_params(method, url)
        if not handler:
            return super().request(
                method,
                url,
                content=content,
                data=data,
                files=files,
                json=json,
                params=params,
                headers=headers,
                cookies=cookies,
                auth=auth,
                follow_redirects=follow_redirects,
                allow_redirects=allow_redirects,
                timeout=timeout,
                extensions=extensions,
            )

        injection_info = self._get_endpoint_injection_info(handler)

        need_mocking_calls = (
            injection_info['installation'] and installation
            or injection_info['installation_admin_client'] and 'installation_id' in path_params
        )

        if not need_mocking_calls:
            return super().request(
                method,
                url,
                content=content,
                data=data,
                files=files,
                json=json,
                params=params,
                headers=headers,
                cookies=cookies,
                auth=auth,
                follow_redirects=follow_redirects,
                allow_redirects=allow_redirects,
                timeout=timeout,
                extensions=extensions,
            )

        mocker = self._get_client_mocker(handler)

        with mocker(self._url_prefix) as mocker, patch.dict(
            os.environ, {'API_KEY': 'ApiKey SU-000:XXX'},
        ):
            if injection_info['installation'] and installation:
                mocker.ns('devops').collection('installations').resource(
                    installation['id'],
                ).get(return_value=installation)
            if injection_info['installation_admin_client'] and 'installation_id' in path_params:
                mocker('devops').services[
                    context.extension_id
                ].installations[
                    path_params['installation_id']
                ]('impersonate').post(
                    return_value={'installation_api_key': 'api-key'},
                )
            return super().request(
                method,
                url,
                content=content,
                data=data,
                files=files,
                json=json,
                params=params,
                headers=headers,
                cookies=cookies,
                auth=auth,
                follow_redirects=follow_redirects,
                allow_redirects=allow_redirects,
                timeout=timeout,
                extensions=extensions,
            )

    def _get_client_mocker(self, handler):
        if inspect.iscoroutinefunction(handler):
            return AsyncConnectClientMocker
        return ConnectClientMocker

    def _get_endpoint_injection_info(self, handler):
        inject_module = asynchronous if inspect.iscoroutinefunction(handler) else synchronous
        signature = inspect.signature(handler)
        info = {
            'extension_client': False,
            'installation_client': False,
            'installation_admin_client': False,
            'installation': False,
        }
        for param in signature.parameters.values():
            if param.default != inspect.Parameter.empty and isinstance(param.default, Depends):
                if param.default.dependency == inject_module.get_installation_client:
                    info['installation_client'] = True
                if param.default.dependency == inject_module.get_installation_admin_client:
                    info['installation_admin_client'] = True
                if param.default.dependency == inject_module.get_installation:
                    info['installation'] = True
        return info

    def _get_endpoint_handler_and_params(self, method, url):
        path = urlparse(url).path
        for route in self.app.router.routes:
            match, child_scope = route.matches({'type': 'http', 'method': method, 'path': path})
            if match == Match.FULL:
                return child_scope['endpoint'], child_scope['path_params']
        return None, None

    def _generate_call_context(self, installation):
        return {
            'extension_id': 'SRVC-0000',
            'environment_id': 'ENV-0000-03',
            'environment_type': 'production',
            'installation_id': installation['id'] if installation else None,
            'user_id': 'UR-000',
            'account_id': 'VA-000',
            'account_role': 'vendor',
            'call_source': 'ui',
            'call_type': 'user',
        }

    def _populate_internal_headers(
        self,
        headers,
        config=None,
        installation=None,
        context=None,
        log_level=None,
    ):

        headers['X-Connect-Logging-Level'] = log_level or 'INFO'
        if config:
            headers['X-Connect-Config'] = json.dumps(config)
        if context.installation_id:
            headers['X-Connect-Installation-id'] = context.installation_id
        headers['X-Connect-Extension-Id'] = context.extension_id
        headers['X-Connect-Environment-Id'] = context.environment_id
        headers['X-Connect-Environment-Type'] = context.environment_type
        headers['X-Connect-User-Id'] = context.user_id
        headers['X-Connect-Account-Id'] = context.account_id
        headers['X-Connect-Account-Role'] = context.account_role
        headers['X-Connect-Call-Source'] = context.call_source
        headers['X-Connect-Call-Type'] = context.call_type
        return headers

    def _get_application(self):
        handlers = {
            ClientError: client_error_exception_handler,
        }
        if hasattr(self._webapp_class, 'get_exception_handlers'):
            handlers = self._webapp_class.get_exception_handlers(handlers)
        app = FastAPI(
            exception_handlers=handlers,
        )

        auth_router, no_auth_router = self._webapp_class.get_routers()
        app.include_router(auth_router, prefix='/api')
        app.include_router(no_auth_router, prefix='/guest')
        app.include_router(no_auth_router, prefix='/unauthorized')

        static_root = self._webapp_class.get_static_root()
        if static_root:
            app.mount('/static', StaticFiles(directory=static_root), name='static')

        if hasattr(self._webapp_class, 'get_middlewares'):
            self._setup_middlewares(app, self._webapp_class.get_middlewares())

        return app

    @staticmethod
    def _setup_middlewares(app, middlewares):
        for middleware in middlewares:
            if inspect.isfunction(middleware):
                app.add_middleware(BaseHTTPMiddleware, dispatch=middleware)
            elif inspect.isclass(middleware):
                app.add_middleware(middleware)
            elif isinstance(middleware, tuple):
                app.add_middleware(middleware[0], **middleware[1])

    def get(self, *args, **kwargs):
        return self.request(
            'GET',
            *args,
            **kwargs,
        )

    def post(self, *args, **kwargs):
        return self.request(
            'POST',
            *args,
            **kwargs,
        )

    def put(self, *args, **kwargs):
        return self.request(
            'PUT',
            *args,
            **kwargs,
        )

    def delete(self, *args, **kwargs):
        return self.request(
            'DELETE',
            *args,
            **kwargs,
        )

    def head(self, *args, **kwargs):
        return self.request(
            'HEAD',
            *args,
            **kwargs,
        )

    def options(self, *args, **kwargs):
        return self.request(
            'OPTIONS',
            *args,
            **kwargs,
        )

    def patch(self, *args, **kwargs):
        return self.request(
            'PATCH',
            *args,
            **kwargs,
        )
